using System;
using System.Collections.Generic;
using System.Text;

namespace Proyecto0
{
    /// <sumary>
    ///  Interface Enumeration al estilo de Java
    /// </sumary>    
    public interface Enumeration
    {
        /// <sumary>
        /// Comprueba si hay ms elementos para iterar
        /// </sumary>
        bool hasMoreElements();

        /// <sumary>
        /// Devuelve el elemento actual y avanza una posicin
        /// </sumary>
        object nextElement();
    }

    /// <sumary>
    /// Se lanza cuando se intenta acceder a una posicin inexistente
    /// </sumary>
    internal class NoSuchElementException : Exception
    {
    }


    /// <sumary>
    ///  Clase interna Nodo
    /// </sumary>
    internal class Nodo
    {
        internal Nodo anterior, siguiente;
        internal object valor;

        public Nodo(object elem, Nodo nodoAnt, Nodo nodoSig)
        {
            valor = elem;
            anterior = nodoAnt;
            siguiente = nodoSig;
        }
    }

    /// <sumary>
    ///  Clase Lista genrica
    /// </sumary>
    public class Lista
    {
        private Nodo cabeza;
        private Nodo cola;
        private int numElem;

        public Lista()
        {
            numElem = 0;
        }

        /// <sumary>
        ///  Aade un elemento a la cabeza de la lista
        /// </sumary>
        public void insertarCabeza(object elem)
        {
            Nodo nodo = new Nodo(elem, null, cabeza);

            if (cabeza != null)
                cabeza.anterior = nodo;
            else
                cola = nodo;

            cabeza = nodo;
            numElem++;
        }
        /// <sumary>
        ///  Aade un elemento a la cola de la lista
        /// </sumary>
        public void insertarCola(object elem)
        {
            Nodo nodo = new Nodo(elem, cola, null);

            if (cola != null)
                cola.siguiente = nodo;
            else
                cabeza = nodo;

            cola = nodo;
            numElem++;
        }

        /// <sumary>
        ///  Extrae un elemento de la cabeza de la lista
        /// </sumary>
        public object extraerCabeza()
        {
            if (cabeza == null)
                return null;

            Nodo nodo = cabeza;
            cabeza = cabeza.siguiente;

            if (cabeza != null)
                cabeza.anterior = null;
            else
                cola = null;

            numElem--;
            return nodo.valor;
        }

        /// <sumary>
        ///  Extrae un elemento de la cola de la lista
        /// </sumary>
        public object extraerCola()
        {
            if (cola == null)
                return null;

            Nodo nodo = cola;
            cola = cola.anterior;

            if (cola != null)
                cola.siguiente = null;
            else
                cabeza = null;

            numElem--;
            return nodo.valor;
        }

        /// <sumary>
        ///  Devuelve true si la lista est vaca
        /// </sumary>
        public bool estaVacia()
        {
            return cabeza == null;
        }


        /// <sumary>
        ///  Devuelve el nmero de elementos en la lista
        /// </sumary>
        public int longitud()
        {
            return numElem;
        }

        /// <sumary>
        ///  Concatena otra lista a la cabeza de la lista
        /// </sumary>
        public void concatenar(Lista otraLista)
        {
            Nodo nodo = otraLista.cabeza;

            while (nodo != null)
            {
                insertarCola(nodo.valor);
                nodo = nodo.siguiente;
            }
        }

        /// <sumary>
        ///  Borra todos los elementos de la lista
        /// </sumary>
        public void borrar()
        {
            cabeza = cola = null;
        }

        /// <sumary>
        ///  Extrae un elemento de la cabeza de la lista sin borrarlo
        /// </sumary>
        public object peekCabeza()
        {
            if (cabeza != null)
                return cabeza.valor;
            else
                return null;
        }

        /// <sumary>
        ///  Extrae un elemento de la cola de la lista sin borrarlo
        /// </sumary>
        public object peekCola()
        {
            if (cola != null)
                return cola.valor;
            else
                return null;
        }


        /// <sumary>
        ///  Devuelve true si la lista contiene el elemento
        /// </sumary>
        public bool contiene(object elem)
        {
            Nodo nodo = cabeza;

            while (nodo != null && !nodo.valor.Equals(elem))
                nodo = nodo.siguiente;

            return nodo != null;
        }

        /// <sumary>
        ///  Devuelve un clon de la lista
        /// </sumary>       
        public object clone()
        {
            Lista temp = new Lista();
            Nodo nodo = cabeza;

            while (nodo != null)
            {
                temp.insertarCola(nodo.valor);
                nodo = nodo.siguiente;
            }

            return temp;
        }

        /// <sumary>
        ///  Crea una cadena que representa la lista
        /// </sumary>
        public override string ToString()
        {
            string temp = "[";
            Nodo nodo = cabeza;

            while (nodo != null)
            {
                temp += nodo.valor.ToString();
                nodo = nodo.siguiente;
                if (nodo != null)
                    temp += ", ";
            }
            temp += "]";

            return temp;
        }

        /// <sumary>
        ///  Iterador Lista
        /// </sumary>
        class Enum : Enumeration
        {
            // Current element
            private Nodo nodo;

            internal Enum(Nodo start)
            {
                nodo = start;
            }

            /// <sumary>
            /// Comprueba si hay ms elementos para iterar
            /// </sumary>
            public bool hasMoreElements()
            {
                return nodo != null;
            }

            /// <sumary>
            /// Devuelve el elemento actual y avanza una posicin
            /// </sumary>
            public object nextElement()
            {
                Object temp;

                if (nodo == null)
                    throw new NoSuchElementException();

                temp = nodo.valor;
                nodo = nodo.siguiente;

                return temp;
            }
        }

        /// <sumary>
        ///  Devuelve un iterador de Lista
        /// </sumary>
        public Enumeration elements()
        {
            return new Enum(cabeza);
        }

    }  

}
